#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.22 16:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt, QEvent, QTimer
from PySide6.QtGui import (
    QAction,
    QIcon,
    QKeySequence,
    QTextCursor,
    QFontMetrics,
    QColor,
)
from PySide6.QtWidgets import QTextEdit, QWidget, QVBoxLayout

from pygpt_net.core.tabs.tab import Tab
from pygpt_net.core.text.finder import Finder
from pygpt_net.ui.widget.element.labels import HelpLabel
from pygpt_net.utils import trans
from .highlight import MarkerHighlighter


class NotepadWidget(QWidget):
    def __init__(self, window=None):
        """
        Notepad

        :param window: main window
        """
        super(NotepadWidget, self).__init__(window)
        self.window = window
        self.id = 1  # assigned in setup
        self.textarea = NotepadOutput(self.window)
        self.window.ui.nodes['tip.output.tab.notepad'] = HelpLabel(trans('tip.output.tab.notepad'), self.window)
        self.opened = False
        self.tab = None

        layout = QVBoxLayout()
        layout.addWidget(self.textarea)
        layout.addWidget(self.window.ui.nodes['tip.output.tab.notepad'])
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        self.setProperty('class', 'layout-notepad')

    def set_tab(self, tab: Tab):
        """
        Set tab

        :param tab: Tab
        """
        self.tab = tab
        self.textarea.set_tab(tab)

    def scroll_to_bottom(self):
        """Scroll down"""
        self.textarea.scroll_to_bottom()

    def setText(self, text: str):
        """
        Set text

        :param text: Text
        """
        self.textarea.setText(text)
        self.textarea.on_update()

    def toPlainText(self) -> str:
        """
        Get plain text

        :return: Plain text
        """
        return self.textarea.toPlainText()

    def on_destroy(self):
        """On destroy"""
        # unregister finder from memory
        self.window.controller.finder.unset(self.textarea.finder)

    def on_delete(self):
        """On delete"""
        self.tab = None  # clear tab reference
        self.textarea.on_delete()  # delete textarea
        self.deleteLater()

class NotepadOutput(QTextEdit):
    ICON_VOLUME = QIcon(":/icons/volume.svg")
    ICON_SAVE = QIcon(":/icons/save.svg")
    ICON_SEARCH = QIcon(":/icons/search.svg")
    ICON_MARK = QIcon(":/icons/edit.svg")
    ICON_UNMARK = QIcon(":/icons/close.svg")

    def __init__(self, window=None):
        """
        Notepad output textarea

        :param window: main window
        """
        super(NotepadOutput, self).__init__(window)
        self.window = window
        self.finder = Finder(window, self)
        self.setAcceptRichText(False)
        self.setStyleSheet(self.window.controller.theme.style('font.chat.output'))
        self.value = self.window.core.config.data['font_size']
        self.max_font_size = 42
        self.min_font_size = 8
        self.id = 1  # assigned in setup
        self.textChanged.connect(self.text_changed)
        self.tab = None
        self.last_scroll_pos = None
        self.installEventFilter(self)
        self.setProperty('class', 'layout-notepad')
        self.initialized = False

        metrics = QFontMetrics(self.font())
        space_width = metrics.horizontalAdvance(" ")
        self.setTabStopDistance(4 * space_width)

        self._vscroll = self.verticalScrollBar()
        self._vscroll.valueChanged.connect(self._on_scrollbar_value_changed)
        self._restore_attempts = 0
        self._pending_scroll_pos = None

        # highlight state (using QSyntaxHighlighter for rendering)
        self._highlights = []  # list of (start, length)

        # timers/slots must be available even if later connections fail
        self._save_timer = QTimer(self)
        self._save_timer.setSingleShot(True)
        self._save_timer.setInterval(400)
        self._save_timer.timeout.connect(self._persist)

        # highlighter
        self._highlighter = MarkerHighlighter(self.document(), self.get_highlights, self.get_highlight_color)

    def on_delete(self):
        """On delete"""
        if self.finder:
            self.finder.disconnect()  # disconnect finder
            self.finder = None  # delete finder
        if self._save_timer.isActive():
            self._save_timer.stop()
        self.deleteLater()

    def showEvent(self, event):
        """On show event"""
        super().showEvent(event)
        self._restore_attempts = 0
        QTimer.singleShot(0, self.restore_scroll_pos)
        self.initialized = True

    def changeEvent(self, event):
        """React to theme/palette changes"""
        if event.type() in (QEvent.PaletteChange, QEvent.ApplicationPaletteChange, QEvent.StyleChange):
            try:
                self._highlighter.rehighlight()
            except Exception:
                pass
        super().changeEvent(event)

    def scroll_to_bottom(self):
        """Scroll to bottom"""
        self.moveCursor(QTextCursor.End)
        self.ensureCursorVisible()
        scroll_bar = self._vscroll
        scroll_bar.setValue(scroll_bar.maximum())

    def eventFilter(self, source, event):
        """
        Focus event filter

        :param source: source
        :param event: event
        """
        if event.type() == QEvent.FocusIn:
            if self.tab is not None:
                col_idx = self.tab.column_idx
                self.window.controller.ui.tabs.on_column_focus(col_idx)
        return super().eventFilter(source, event)

    def set_tab(self, tab: Tab):
        """
        Set tab

        :param tab: Tab
        """
        self.tab = tab

    def setText(self, text: str):
        """
        Set text

        :param text: Text
        """
        if self.toPlainText() == text:
            return
        self.setPlainText(text)
        self._highlighter.rehighlight()  # refresh highlighting on new content

    def text_changed(self):
        """On text changed"""
        if not self.window.core.notepad.locked:
            if self.finder is not None:
                self.finder.text_changed()
            self.last_scroll_pos = self._vscroll.value()
            if self.initialized and not self.toPlainText():
                QTimer.singleShot(0, lambda: self.clear_highlights(persist=False))  # if empty, reset highlights
            self.schedule_save()

    def _on_scrollbar_value_changed(self, value: int):
        """
        On scrollbar value changed

        :param value: New value
        """
        if not self.window.core.notepad.locked:
            self.last_scroll_pos = value
            self.schedule_save()

    def is_initialized(self) -> bool:
        """
        Check if initialized

        :return: True if initialized
        """
        return self.initialized

    def restore_scroll_pos(self):
        """Restore last scroll position"""
        if self.window.core.notepad.locked:
            QTimer.singleShot(25, self.restore_scroll_pos)
            return
        if not self.initialized:
            return
        if self._pending_scroll_pos is not None:
            self.last_scroll_pos = self._pending_scroll_pos
        if self.last_scroll_pos is None:
            return
        scroll_bar = self._vscroll
        current_max = scroll_bar.maximum()
        if current_max == 0:
            return  # nothing to scroll
        if self.last_scroll_pos > current_max:
            if self._restore_attempts < 30:
                self._restore_attempts += 1
                QTimer.singleShot(16, self.restore_scroll_pos)
            else:
                scroll_bar.setValue(current_max)
        else:
            self.window.core.notepad.locked = True
            scroll_bar.setValue(self.last_scroll_pos)
            if self._pending_scroll_pos is not None:
                self._pending_scroll_pos = None
            self.window.core.notepad.locked = False

    def set_scroll_pos(self, pos: int):
        """
        Set scroll position

        :param pos: Scroll position
        """
        self._pending_scroll_pos = pos
        self.last_scroll_pos = pos

    def get_scroll_pos(self) -> int:
        """
        Get scroll position

        :return: Scroll position
        """
        return self._vscroll.value()

    def schedule_save(self):
        """Schedule save of notepad content"""
        try:
            self._save_timer.start()
        except Exception:
            pass

    def contextMenuEvent(self, event):
        """
        Context menu event

        :param event: Event
        """
        menu = self.createStandardContextMenu()
        cursor = self.textCursor()
        selected_text = cursor.selectedText()
        has_selection = bool(selected_text)

        # Mark / Unmark actions for selection
        if has_selection:
            start = min(cursor.selectionStart(), cursor.selectionEnd())
            end = max(cursor.selectionStart(), cursor.selectionEnd())
            overlap = self._selection_overlaps_any_highlight(start, end)

            action_mark = QAction(self.ICON_MARK, trans("action.mark"), self)
            action_mark.triggered.connect(self.mark_selection)
            menu.addAction(action_mark)

            action_unmark = QAction(self.ICON_UNMARK, trans("action.unmark"), self)
            action_unmark.setEnabled(overlap)
            action_unmark.triggered.connect(self.unmark_selection)
            menu.addAction(action_unmark)

        if selected_text:
            plain_text = cursor.selection().toPlainText()

            action = QAction(self.ICON_VOLUME, trans('text.context_menu.audio.read'), self)
            action.triggered.connect(self.audio_read_selection)
            menu.addAction(action)

            excluded_id = f"notepad_id_{self.id}"
            copy_to_menu = self.window.ui.context_menu.get_copy_to_menu(menu, selected_text, excluded=[excluded_id])
            menu.addMenu(copy_to_menu)

            action = QAction(self.ICON_SAVE, trans('action.save_selection_as'), self)
            action.triggered.connect(
                lambda: self.window.controller.chat.common.save_text(plain_text))
            menu.addAction(action)
        else:
            action = QAction(self.ICON_SAVE, trans('action.save_as'), self)
            action.triggered.connect(
                lambda: self.window.controller.chat.common.save_text(self.toPlainText()))
            menu.addAction(action)

        # Add zoom submenu
        zoom_menu = self.window.ui.context_menu.get_zoom_menu(self, "font_size", self.value, self.on_zoom_changed)
        menu.addMenu(zoom_menu)

        action = QAction(self.ICON_SEARCH, trans('text.context_menu.find'), self)
        action.triggered.connect(self.find_open)
        action.setShortcut(QKeySequence("Ctrl+F"))
        menu.addAction(action)

        menu.exec_(event.globalPos())

    def audio_read_selection(self):
        """
        Read selected text (audio)
        """
        self.window.controller.audio.read_text(self.textCursor().selectedText())

    def find_open(self):
        """Open find dialog"""
        self.window.controller.finder.open(self.finder)

    def on_update(self):
        """On content update"""
        self.finder.clear()  # clear finder

    def on_zoom_changed(self, value: int):
        """
        On font size changed

        :param value: New font size
        """
        self.value = value
        self.window.core.config.data['font_size'] = value
        self.window.core.config.save()
        option = self.window.controller.settings.editor.get_option('font_size')
        option['value'] = self.value
        self.window.controller.config.apply(
                parent_id='config',
                key='font_size',
                option=option,
        )
        self.window.controller.ui.update_font_size()
        self.last_scroll_pos = self._vscroll.value()

    def keyPressEvent(self, e):
        """
        Key press event

        :param e: Event
        """
        if e.key() == Qt.Key_F and e.modifiers() & Qt.ControlModifier:
            self.find_open()
        else:
            self.finder.clear(restore=True, to_end=False)
            super(NotepadOutput, self).keyPressEvent(e)

    def wheelEvent(self, event):
        """
        Wheel event: set font size

        :param event: Event
        """
        if event.modifiers() & Qt.ControlModifier:
            delta = event.angleDelta().y()
            if delta > 0:
                if self.value < self.max_font_size:
                    self.value += 1
                else:
                    return
            else:
                if self.value > self.min_font_size:
                    self.value -= 1
                else:
                    return

            self.window.core.config.data['font_size'] = self.value
            self.window.core.config.save()
            option = self.window.controller.settings.editor.get_option('font_size')
            option['value'] = self.value
            self.window.controller.config.apply(
                parent_id='config',
                key='font_size',
                option=option,
            )
            self.window.controller.ui.update_font_size()
            event.accept()
            self.last_scroll_pos = self._vscroll.value()
        else:
            super(NotepadOutput, self).wheelEvent(event)
            self.last_scroll_pos = self._vscroll.value()

    def focusInEvent(self, e):
        """
        Focus in event

        :param e: focus event
        """
        self.window.controller.finder.focus_in(self.finder)
        super(NotepadOutput, self).focusInEvent(e)

    def focusOutEvent(self, e):
        """
        Focus out event

        :param e: focus event
        """
        super(NotepadOutput, self).focusOutEvent(e)
        self.window.controller.finder.focus_out(self.finder)

    # ==== Marking / Highlights API ====

    def mark_selection(self):
        """Apply highlight to current selection"""
        cursor = self.textCursor()
        if not cursor.hasSelection():
            return
        start = min(cursor.selectionStart(), cursor.selectionEnd())
        end = max(cursor.selectionStart(), cursor.selectionEnd())
        self._add_highlight((start, end - start))
        self._highlighter.rehighlight()
        self._persist()

    def unmark_selection(self):
        """Remove highlight from current selection"""
        cursor = self.textCursor()
        if not cursor.hasSelection():
            return
        start = min(cursor.selectionStart(), cursor.selectionEnd())
        end = max(cursor.selectionStart(), cursor.selectionEnd())
        self._remove_range_from_highlights(start, end - start)
        self._highlighter.rehighlight()
        self._persist()

    def get_highlights(self):
        """Return current highlights as list of (start, length)"""
        return list(self._highlights)

    def set_highlights(self, highlights):
        """Set highlights and repaint"""
        self._highlights = self._merge_ranges(self._sanitize_ranges(highlights))
        self._highlighter.rehighlight()

    def clear_highlights(self, persist: bool = True):
        """Clear all highlights"""
        self._highlights = []
        self._highlighter.rehighlight()
        if persist:
            self._persist()

    # ==== Highlight colors / theme ====

    def get_highlight_color(self):
        """
        Return (text_color, background_color) for highlights based on current theme.
        """
        is_dark = self._is_dark_theme()
        if is_dark:
            text_color = QColor(0, 0, 0)
            bg_color = QColor(255, 255, 0)  # yellow
        else:
            text_color = QColor(0, 0, 0)
            bg_color = QColor(255, 255, 0)  # yellow
        return text_color, bg_color

    def apply_highlight_theme(self):
        """
        Public method to refresh highlight colors. Can be called by theme controller
        after theme switch.
        """
        try:
            self._highlighter.rehighlight()
        except Exception:
            pass

    def _is_dark_theme(self) -> bool:
        """
        Get whether current theme is dark
        """
        return self.window.controller.theme.is_dark_theme()

    # ==== Internal highlight helpers ====

    def _persist(self):
        """Persist notepad state"""
        if self._save_timer.isActive():
            self._save_timer.stop()
        try:
            self.window.controller.notepad.save(self.id)  # save content + marking
        except Exception as e:
            print(e)

    def _sanitize_ranges(self, ranges):
        """Sanitize ranges to (start>=0, length>0) integers"""
        out = []
        for r in ranges or []:
            try:
                s = int(r[0])
                l = int(r[1])
            except Exception:
                continue
            if s < 0 or l <= 0:
                continue
            out.append((s, l))
        out.sort(key=lambda x: x[0])
        return out

    def _merge_ranges(self, ranges):
        """Merge overlapping/adjacent ranges"""
        if not ranges:
            return []
        merged = []
        for s, l in ranges:
            if not merged:
                merged.append([s, s + l])
                continue
            ps, pe = merged[-1]
            se = s + l
            if s <= pe:
                merged[-1][1] = max(pe, se)
            else:
                merged.append([s, se])
        return [(s, e - s) for s, e in merged]

    def _add_highlight(self, rng):
        """Add a highlight range and merge"""
        s, l = int(rng[0]), int(rng[1])
        if l <= 0:
            return
        self._highlights.append((s, l))
        self._highlights = self._merge_ranges(self._sanitize_ranges(self._highlights))
        self.schedule_save()

    def _remove_range_from_highlights(self, s, l):
        """Subtract a range from all highlights"""
        if l <= 0:
            return
        start = s
        end = s + l
        result = []
        for hs, hl in self._highlights:
            he = hs + hl
            if he <= start or hs >= end:
                result.append((hs, hl))
                continue
            if hs < start:
                result.append((hs, start - hs))
            if he > end:
                result.append((end, he - end))
        self._highlights = self._merge_ranges(self._sanitize_ranges(result))
        self.schedule_save()

    def _selection_overlaps_any_highlight(self, s, e):
        """Check if selection overlaps any highlight"""
        for hs, hl in self._highlights:
            he = hs + hl
            if not (he <= s or hs >= e):
                return True
        return False


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.21 20:00:00                  #
# ================================================== #

from PySide6.QtGui import (
    QTextCharFormat,
    QSyntaxHighlighter,
)


class MarkerHighlighter(QSyntaxHighlighter):
    def __init__(self, document, ranges_provider, colors_provider):
        """
        Syntax highlighter for marking ranges.

        :param document: QTextDocument
        :param ranges_provider: callable returning list of (start, length)
        :param colors_provider: callable returning (text_color: QColor|None, bg_color: QColor|None)
        """
        super().__init__(document)
        self._ranges_provider = ranges_provider
        self._colors_provider = colors_provider

    def set_colors_provider(self, colors_provider):
        """Set a new provider for highlight colors"""
        self._colors_provider = colors_provider
        self.rehighlight()

    def highlightBlock(self, text: str):
        """Apply formatting to ranges intersecting current block"""
        ranges = self._ranges_provider() or []
        if not ranges:
            return

        text_color, bg_color = (None, None)
        try:
            text_color, bg_color = self._colors_provider()
        except Exception:
            pass

        fmt = QTextCharFormat()
        if bg_color is not None:
            fmt.setBackground(bg_color)
        if text_color is not None:
            fmt.setForeground(text_color)

        block_start = self.currentBlock().position()
        block_len = len(text)
        for s, l in ranges:
            if l <= 0:
                continue
            rel_start = s - block_start
            rel_end = (s + l) - block_start
            if rel_end <= 0 or rel_start >= block_len:
                continue
            rel_start = max(0, rel_start)
            length = min(block_len - rel_start, rel_end - rel_start)
            if length > 0:
                self.setFormat(rel_start, length, fmt)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.22 04:00:00                  #
# ================================================== #

from typing import Optional, Any

from PySide6.QtCore import Slot, QObject

from pygpt_net.core.bridge import BridgeContext
from pygpt_net.core.events import RenderEvent
from pygpt_net.core.types import MODE_ASSISTANT
from pygpt_net.item.ctx import CtxItem

from .stream_worker import StreamWorker

class Stream(QObject):
    def __init__(self, window=None):
        """
        Stream controller

        :param window: Window instance
        """
        super().__init__()
        self.window = window
        self.instance = None  # cached get renderer instance method
        self.pids = {} # {pid -> {data}}, workers are tracked per PID

    def get_pid_by_ctx(self, ctx: CtxItem) -> Optional[int]:
        """
        Get process ID by context item

        :param ctx: Context item
        :return: Process ID or None
        """
        if ctx and ctx.meta:
            return self.window.core.ctx.output.get_pid(ctx.meta)
        return None

    def get_pid_ids(self) -> list[int]:
        """
        Get all active process IDs

        :return: List of process IDs
        """
        return list(self.pids.keys())

    def append(
            self,
            ctx: CtxItem,
            mode: str = None,
            is_response: bool = False,
            reply: str = False,
            internal: bool = False,
            context: Optional[BridgeContext] = None,
            extra: Optional[dict] = None
    ):
        """
        Asynchronous append of stream worker to the thread.

        :param ctx: Context item
        :param mode: Mode of operation (e.g., MODE_ASSISTANT)
        :param is_response: Whether this is a response stream
        :param reply: Reply identifier
        :param internal: Whether this is an internal stream
        :param context: Optional BridgeContext for additional context
        :param extra: Additional data to pass to the stream
        """
        pid = self.get_pid_by_ctx(ctx)
        if pid is None:
            return  # abort streaming if no PID found
        if pid not in self.pids:
            self.pids[pid] = {}

        pid_data = self.pids[pid]
        pid_data["ctx"] = ctx
        pid_data["mode"] = mode
        pid_data["is_response"] = is_response
        pid_data["reply"] = reply
        pid_data["internal"] = internal
        pid_data["context"] = context
        pid_data["extra"] = extra if extra is not None else {}

        # cache the get renderer instance method
        if self.instance is None:
            self.instance = self.window.controller.chat.render.instance # callable

        worker = StreamWorker(ctx, self.window)
        worker.stream = ctx.stream
        worker.signals.eventReady.connect(self.handleEvent)
        worker.signals.errorOccurred.connect(self.handleError)
        worker.signals.end.connect(self.handleEnd)
        worker.signals.chunk.connect(self.handleChunk)
        ctx.stream = None # clear reference to generator

        pid_data["worker"] = worker # keep reference to avoid GC, per PID
        self.window.core.debug.info(f"[chat] Stream begin... PID={pid}")
        self.window.threadpool.start(worker)

    @Slot(object)
    def handleEnd(self, ctx: CtxItem):
        """
        Slot for handling end of stream

        :param ctx: Context item
        """
        pid = self.get_pid_by_ctx(ctx)
        if pid is None or pid not in self.pids:
            return  # abort if no PID found or not tracked
        pid_data = self.pids[pid]

        controller = self.window.controller
        controller.ui.update_tokens()
        mode = pid_data["mode"]

        data = {
            "meta": pid_data["ctx"].meta,
            "ctx": pid_data["ctx"]
        }
        event = RenderEvent(RenderEvent.STREAM_END, data)
        self.window.dispatch(event)
        controller.chat.output.handle_after(
            ctx=ctx,
            mode=mode,
            stream=True,
        )

        if mode == MODE_ASSISTANT:
            controller.assistant.threads.handle_output_message_after_stream(ctx)
        else:
            if pid_data["is_response"]:
                controller.chat.response.post_handle(
                    ctx=ctx,
                    mode=mode,
                    stream=True,
                    reply=pid_data["reply"],
                    internal=pid_data["internal"],
                )

        pid_data["worker"] = None  # release worker reference
        del self.pids[pid]  # remove PID tracking

    @Slot(object, str, bool)
    def handleChunk(
            self,
            ctx: CtxItem,
            chunk: str,
            begin: bool = False
    ):
        """
        Handle a chunk of data in the stream

        :param ctx: Context item
        :param chunk: Chunk of data
        :param begin: Whether this is the beginning of the stream
        """
        # direct call to the renderer to avoid overhead of event queue
        self.instance().append_chunk(
            ctx.meta,
            ctx,
            chunk,
            begin,
        )
        chunk = None  # free reference

    @Slot(object)
    def handleEvent(self, event):
        """
        Slot for handling stream events

        :param event: RenderEvent
        """
        self.window.dispatch(event)

    @Slot(object, object)
    def handleError(self, ctx: CtxItem, error: Any):
        """
        Slot for handling stream errors

        :param ctx: Context item
        :param error: Exception or error message
        """
        pid = self.get_pid_by_ctx(ctx)
        if pid is None or pid not in self.pids:
            return  # abort if no PID found or not tracked
        pid_data = self.pids[pid]

        self.window.core.debug.log(error)
        if pid_data["is_response"]:
            if not isinstance(pid_data["extra"], dict):
                pid_data["extra"] = {}
            pid_data["extra"]["error"] = error
            self.window.controller.chat.response.failed(pid_data["context"], pid_data["extra"])
            self.window.controller.chat.response.post_handle(
                ctx=pid_data["ctx"],
                mode=pid_data["mode"],
                stream=True,
                reply=pid_data["reply"],
                internal=pid_data["internal"],
            )
        # TODO: remove PID from tracking on error?

    def log(self, data: object):
        """
        Log data to the debug console

        :param data: object to log
        """
        self.window.core.debug.info(data)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.22 16:00:00                  #
# ================================================== #

from typing import Optional, Tuple

from PySide6.QtCore import QTimer
from PySide6.QtGui import QTextCursor

from pygpt_net.core.tabs.tab import Tab
from pygpt_net.item.notepad import NotepadItem
from pygpt_net.ui.widget.tabs.body import TabBody
from pygpt_net.ui.widget.textarea.notepad import NotepadWidget
from pygpt_net.utils import trans


class Notepad:
    def __init__(self, window=None):
        """
        Notepad controller

        :param window: Window instance
        """
        self.window = window
        self.opened_once = False
        self.opened_idx = set()

    def get_next_suffix(self) -> int:
        """
        Get next notepad suffix

        :return: next notepad suffix
        """
        tabs_core = self.window.core.tabs
        tabs = tabs_core.get_tabs_by_type(Tab.TAB_NOTEPAD)
        idx = tabs_core.count_by_type(Tab.TAB_NOTEPAD)
        for tab in tabs:
            title = tab.title or ""
            try:
                num = int(title.rsplit(" ", 1)[-1])
                if num > idx:
                    idx = num
            except (ValueError, IndexError):
                continue
        return idx + 1

    def is_opened(self, idx: int) -> bool:
        """
        Get next available notepad index

        :return: next available notepad index
        """
        tabs = self.window.core.tabs.get_tabs_by_type(Tab.TAB_NOTEPAD)
        return any(tab.data_id == idx for tab in tabs)

    def create(
            self,
            idx: Optional[int] = None,
            tab: Optional[Tab] = None,
            restore: bool = False
    ) -> Tuple[TabBody, int, int]:
        """
        Create notepad widget

        :param idx: notepad idx
        :param tab: existing tab to use (optional)
        :param restore: whether to restore existing notepad (default: False)
        :return: notepad widget (TabBody)
        """
        tabs_core = self.window.core.tabs
        if not restore:
            existing_tabs = tabs_core.get_tabs_by_type(Tab.TAB_NOTEPAD)
            used_ids = {t.data_id for t in existing_tabs}
            if idx is None:
                idx = 1
            while idx in used_ids:
                idx += 1

        suffix = self.get_next_suffix()

        data_id = idx
        widget = NotepadWidget(self.window)
        widget.id = idx
        widget.textarea.id = idx
        self.window.ui.notepad[data_id] = widget

        title = trans('output.tab.notepad') + " " + str(suffix)
        if tab:
            if not tab.title:
                tab.title = title

        child = tabs_core.from_widget(widget)
        child.on_delete = self.on_delete
        return child, idx, data_id

    def on_delete(self, body: TabBody):
        """
        On delete notepad tab

        :param body: TabBody instance
        """
        tab = body.owner
        if tab.type != Tab.TAB_NOTEPAD:
            return
        idx = tab.data_id
        if idx in self.window.ui.notepad:
            self.window.ui.notepad[idx].on_delete()
            del self.window.ui.notepad[idx]
        if idx in self.opened_idx:
            self.opened_idx.discard(idx)
        self.update()

    def load(self):
        """Load all notepads"""
        core_notepad = self.window.core.notepad
        prev_locked = core_notepad.locked
        core_notepad.locked = True  # block any autosave during restore
        try:
            core_notepad.load_all()
            items = core_notepad.get_all()
            num_notepads = self.get_num_notepads()

            if len(items) == 0:
                if num_notepads > 0:
                    for idx in range(1, num_notepads + 1):
                        item = NotepadItem()
                        item.idx = idx
                        items[idx] = item

            if num_notepads > 0:
                for idx, item in items.items():
                    widget = self.window.ui.notepad.get(idx)
                    if widget is not None:
                        widget.setText(item.content)
                        highlights = item.highlights
                        widget.textarea.set_highlights(highlights)
                        if item.scroll_pos is not None and item.scroll_pos != -1:
                            widget.textarea.set_scroll_pos(item.scroll_pos)
        finally:
            core_notepad.locked = prev_locked  # restore previous lock state

    def get_notepad_name(self, idx: int):
        """
        Get notepad name

        :param idx: notepad idx
        :return: notepad name
        """
        num = self.get_num_notepads()
        if num > 1:
            title = trans('text.context_menu.copy_to.notepad') + ' ' + str(idx)
        else:
            title = trans('text.context_menu.copy_to.notepad')
        item = self.window.core.notepad.get_by_id(idx)
        if item is None:
            return None
        if item.initialized and item.title is not None and len(item.title) > 0:
            title = item.title
        return title

    def save(self, idx: int):
        """
        Save notepad contents

        :param idx: notepad idx
        """
        core_notepad = self.window.core.notepad
        item = core_notepad.get_by_id(idx)
        if item is None:
            item = NotepadItem()
            item.idx = idx
            core_notepad.items[idx] = item

        widget = self.window.ui.notepad.get(idx)
        if widget is not None:
            if not widget.textarea.is_initialized():
                return # do not save uninitialized notepads
            text = widget.toPlainText()
            if item.content != text:
                item.content = text
                self.save_state(idx, persist=False)
                core_notepad.update(item)
            else:
                self.save_state(idx)
            self.update()

    def save_state(self, idx: int, persist: bool = True):
        """
        Save highlights for given notepad idx into item.highlights
        """
        widget = self.window.ui.notepad.get(idx)
        if widget is None:
            return
        highlights = widget.textarea.get_highlights()
        scroll_pos = widget.textarea.get_scroll_pos()
        core_notepad = self.window.core.notepad
        item = core_notepad.get_by_id(idx)
        if item is None or core_notepad.locked:
            return
        item.highlights = highlights
        if widget.textarea.is_initialized():
            item.scroll_pos = scroll_pos
        if persist:
            core_notepad.update(item)

    def save_all(self):
        """Save all notepads contents"""
        items = self.window.core.notepad.get_all()
        num_notepads = self.get_num_notepads()
        if num_notepads > 0:
            ui_notepad = self.window.ui.notepad
            for tab in self.window.core.tabs.get_tabs_by_type(Tab.TAB_NOTEPAD):
                idx = tab.data_id
                widget = ui_notepad.get(idx)
                if widget is None:
                    continue
                if idx in items:
                    prev_content = str(items[idx].content)
                    text = widget.toPlainText()
                    if prev_content != text:
                        items[idx].content = text
                        self.save_state(idx, persist=False)  # persist highlights for each notepad
                        self.window.core.notepad.update(items[idx])
        self.update()

    def setup(self):
        """Setup all notepads"""
        self.load()

    def append_text(self, text: str, idx: int):
        """
        Append text to notepad

        :param text: text to append
        :param idx: notepad idx
        """
        widget = self.window.ui.notepad.get(idx)
        if widget is None:
            return
        dt = ""
        prev_text = widget.toPlainText()
        need_nl = prev_text.strip() != ""
        textarea = widget.textarea
        cursor = textarea.textCursor()
        cursor.movePosition(QTextCursor.End)
        to_insert = (("\n" if need_nl else "") + dt + text.strip())
        cursor.insertText(to_insert)
        textarea.setTextCursor(cursor)
        self.save(idx)

    def get_num_notepads(self) -> int:
        """
        Get number of notepads

        :return: number of notepads
        """
        return self.window.core.tabs.count_by_type(Tab.TAB_NOTEPAD)

    def get_current_active(self) -> int:
        """
        Get current notepad idx

        :return: current notepad index
        """
        tabs_ctrl = self.window.controller.ui.tabs
        if self.is_active():
            tab = tabs_ctrl.get_current_tab()
            if tab is not None:
                return tab.data_id
        return 1

    def is_active(self) -> bool:
        """
        Check if notepad tab is active

        :return: True if notepad tab is active
        """
        return self.window.controller.ui.tabs.get_current_type() == Tab.TAB_NOTEPAD

    def open(self):
        """Open notepad"""
        if self.get_num_notepads() == 0:
            return
        if self.window.controller.ui.tabs.get_current_type() != Tab.TAB_NOTEPAD:
            idx = self.window.core.tabs.get_min_idx_by_type(Tab.TAB_NOTEPAD)
            if idx is not None:
                tabs = self.window.ui.layout.get_active_tabs()
                tabs.setCurrentIndex(idx)
        self.window.activateWindow()

    def update(self):
        """Update notepads UI"""
        pass

    def reload(self):
        """Reload notepads"""
        self.window.core.notepad.locked = True
        self.window.core.notepad.reset()
        self.load()
        self.window.core.notepad.locked = False

    def switch_to_tab(self, idx: Optional[int] = None):
        """
        Switch to notepad tab

        :param idx: notepad idx
        """
        if idx is None:
            idx = self.get_first_notepad_tab_idx()
        tabs = self.window.ui.layout.get_active_tabs()
        tab = self.window.core.tabs.get_tab_by_index(idx)
        if tab is not None:
            tabs.setCurrentIndex(idx)
        else:
            tabs.setCurrentIndex(self.get_first_notepad_tab_idx())

    def get_first_notepad_tab_idx(self) -> int:
        """
        Get first notepad tab index

        :return: first notepad tab index
        """
        return self.window.core.tabs.get_min_idx_by_type(Tab.TAB_NOTEPAD)

    def get_current_notepad_text(self) -> str:
        """
        Get current notepad text

        :return: current notepad text
        """
        idx = self.get_current_active()
        widget = self.window.ui.notepad.get(idx)
        return widget.toPlainText() if widget is not None else ""

    def get_notepad_text(self, idx: int) -> str:
        """
        Get notepad text

        :param idx: notepad index
        :return: notepad text
        """
        widget = self.window.ui.notepad.get(idx)
        return widget.toPlainText() if widget is not None else ""

    def clear(self, idx: int) -> bool:
        """
        Clear notepad contents

        :param idx: notepad idx
        """
        widget = self.window.ui.notepad.get(idx)
        if widget is not None:
            widget.textarea.clear()
            widget.textarea.clear_highlights()
            self.save(idx)
            return True
        return False

    def on_open(
            self,
            tab_idx: int,
            column_idx: int = 0
    ):
        """
        On open notepad tab

        :param tab_idx: current tab idx
        :param column_idx: column idx
        """
        tab = self.window.controller.ui.tabs.get_current_tab()
        if tab is None:
            return
        if tab.type == Tab.TAB_NOTEPAD:
            idx = tab.data_id
            widget = self.window.ui.notepad.get(idx)
            if widget is None:
                return
            # only autoscroll brand-new notepad once, if there is no saved scroll position
            if idx not in self.opened_idx:
                item = self.window.core.notepad.get_by_id(idx)
                has_saved_pos = item is not None and item.scroll_pos not in (None, -1)
                if not has_saved_pos:
                    QTimer.singleShot(0, widget.scroll_to_bottom)
            if not widget.opened:
                widget.opened = True
            if idx not in self.opened_idx:
                self.opened_idx.add(idx)

    def focus_opened(self, tab = None):
        """Focus opened notepad"""
        if tab is None:
            tab = self.window.controller.ui.tabs.get_current_tab()
        if tab is not None and tab.type == Tab.TAB_NOTEPAD:
            widget = self.window.ui.notepad.get(tab.data_id)
            if widget is not None:
                QTimer.singleShot(100, widget.textarea.setFocus)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.15 01:00:00                  #
# ================================================== #

from typing import List, Dict, Optional

from llama_index.core.base.embeddings.base import BaseEmbedding

from pygpt_net.core.types import (
    MODE_LLAMA_INDEX,
)
from llama_index.core.llms.llm import BaseLLM as LlamaBaseLLM

from pygpt_net.provider.llms.base import BaseLLM
from pygpt_net.item.model import ModelItem


class DeepseekApiLLM(BaseLLM):
    def __init__(self, *args, **kwargs):
        super(DeepseekApiLLM, self).__init__(*args, **kwargs)
        self.id = "deepseek_api"
        self.name = "Deepseek API"
        self.type = [MODE_LLAMA_INDEX, "embeddings"]

    def llama(
            self,
            window,
            model: ModelItem,
            stream: bool = False
    ) -> LlamaBaseLLM:
        """
        Return LLM provider instance for llama

        :param window: window instance
        :param model: model instance
        :param stream: stream mode
        :return: LLM provider instance
        """
        from llama_index.llms.deepseek import DeepSeek
        args = self.parse_args(model.llama_index, window)
        if "model" not in args:
            args["model"] = model.id
        if "api_key" not in args or args["api_key"] == "":
            args["api_key"] = window.core.config.get("api_key_deepseek", "")
        args = self.inject_llamaindex_http_clients(args, window.core.config)
        return DeepSeek(**args)

    def get_embeddings_model(
            self,
            window,
            config: Optional[List[Dict]] = None
    ) -> BaseEmbedding:
        """
        Return provider instance for embeddings

        :param window: window instance
        :param config: config keyword arguments list
        :return: Embedding provider instance
        """
        from .voyage import VoyageEmbeddingWithProxy
        args = {}
        if config is not None:
            args = self.parse_args({
                "args": config,
            }, window)
        if "api_key" in args:
            args["voyage_api_key"] = args.pop("api_key")
        if "voyage_api_key" not in args or args["voyage_api_key"] == "":
            args["voyage_api_key"] = window.core.config.get("api_key_voyage", "")
        if "model" in args and "model_name" not in args:
            args["model_name"] = args.pop("model")

        timeout = window.core.config.get("api_native_voyage.timeout")
        max_retries = window.core.config.get("api_native_voyage.max_retries")
        proxy = window.core.config.get("api_proxy")
        if not window.core.config.get("api_proxy.enabled", False):
            proxy = ""
        return VoyageEmbeddingWithProxy(
            **args,
            proxy=proxy,
            timeout=timeout,
            max_retries=max_retries,
        )

    def get_models(
            self,
            window,
    ) -> List[Dict]:
        """
        Return list of models for the provider

        :param window: window instance
        :return: list of models
        """
        items = []
        client = self.get_client(window)
        models_list = client.models.list()
        if models_list.data:
            for item in models_list.data:
                items.append({
                    "id": item.id,
                    "name": item.id,
                })
        return items



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.05 23:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QApplication
from PySide6.QtCore import QSize, QPoint, QEvent, Qt
from PySide6.QtWidgets import QDialog


class BaseDialog(QDialog):
    def __init__(self, window=None, id=None):
        """
        Base dialog

        :param window: main window
        :param id: dialog id
        """
        super(BaseDialog, self).__init__(window)
        self.window = window
        self.id = id
        self.shared_id = None
        self.disable_geometry_store = False
        self.setWindowFlags(
            self.windowFlags() | Qt.WindowMaximizeButtonHint
        )

    def _get_id(self):
        """
        Get dialog id

        :return: dialog id
        """
        return self.shared_id if self.shared_id is not None else self.id

    def showEvent(self, event):
        """
        Event called when the dialog is shown

        :param event: show event
        """
        super(BaseDialog, self).showEvent(event)
        if event.type() == QEvent.Show:
            self.restore_geometry()

    def closeEvent(self, event):
        """
        Event called when the dialog is closed

        :param event: close event
        """
        self.save_geometry()
        super(BaseDialog, self).closeEvent(event)

    def store_geometry_enabled(self) -> bool:
        """
        Check if the geometry store is enabled

        :return: True if enabled
        """
        if self.disable_geometry_store:
            return False

        if self.window is None or self._get_id() is None:
            return False

        config = self.window.core.config
        if not config.has("layout.dialog.geometry.store") \
                or not config.get("layout.dialog.geometry.store"):
            return False

        return True

    def save_geometry(self):
        """Save dialog geometry"""
        config = self.window.core.config
        id = self._get_id()
        item = {
            "size": [self.size().width(), self.size().height()],
            "pos": [self.pos().x(), self.pos().y()]
        }
        data = {}
        if self.store_geometry_enabled():
            data = config.get_session("layout.dialog.geometry", {})

        if not isinstance(data, dict):
            data = {}
        data[id] = item

        if self.store_geometry_enabled():
            config.set_session("layout.dialog.geometry", data)

    def restore_geometry(self):
        """Restore dialog geometry"""
        # get available screen geometry
        id = self._get_id()
        screen = QApplication.primaryScreen()
        available_geometry = screen.availableGeometry()
        config = self.window.core.config

        data = {}
        if self.store_geometry_enabled():
            data = config.get_session("layout.dialog.geometry", {})

        item = data.get(id, {})
        if isinstance(item, dict) and "size" in item and "pos" in item:
            width, height = item["size"]
            x, y = item["pos"]
            width = min(width, available_geometry.width()) - 20
            height = min(height, available_geometry.height()) - 20
            size = QSize(width, height)

            # adjust position
            x = max(min(x, available_geometry.right() - width), available_geometry.left())
            y = max(min(y, available_geometry.bottom() - height), available_geometry.top())

            pos = QPoint(x, y)
            self.resize(size)
            self.move(pos)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.20 21:00:00                  #
# ================================================== #

from PySide6 import QtCore
from PySide6.QtWidgets import QLineEdit, QApplication


class ConsoleInput(QLineEdit):
    def __init__(self, window=None):
        """
        Console input

        :param window: main window
        """
        super(ConsoleInput, self).__init__(window)
        self.window = window
        self.setPlaceholderText("Console... Type your command here")
        self.setProperty('class', 'text-editor')
        self.setFocus()

        # command history storage
        self._history = []
        self._history_index = 0  # points to position after the last item
        self._buffer_before_history = ""  # stores current line when entering history navigation
        self._in_history_mode = False

        # supported commands for TAB auto-completion
        self._commands = []

    def set_commands(self, commands):
        """
        Set supported commands for auto-completion.

        :param commands: list of command strings
        """
        if isinstance(commands, (list, tuple)):
            self._commands = list(commands)

    def add_to_history(self, cmd: str):
        """
        Add command to history.

        :param cmd: command string
        """
        if not cmd:
            return
        # avoid adding the same command twice in a row
        if len(self._history) == 0 or self._history[-1] != cmd:
            self._history.append(cmd)
        # reset navigation state
        self._history_index = len(self._history)
        self._in_history_mode = False
        self._buffer_before_history = ""

    def _history_prev(self):
        """Navigate to previous history item (Up)."""
        if not self._history:
            QApplication.beep()
            return
        if not self._in_history_mode:
            self._buffer_before_history = self.text()
            self._in_history_mode = True
        if self._history_index > 0:
            self._history_index -= 1
            self.setText(self._history[self._history_index])
            self.end(False)
        else:
            QApplication.beep()

    def _history_next(self):
        """Navigate to next history item (Down)."""
        if not self._history:
            QApplication.beep()
            return
        if not self._in_history_mode:
            QApplication.beep()
            return
        if self._history_index < len(self._history) - 1:
            self._history_index += 1
            self.setText(self._history[self._history_index])
            self.end(False)
        else:
            # move past the last item: restore buffer and exit history mode
            self._history_index = len(self._history)
            self.setText(self._buffer_before_history)
            self.end(False)
            self._in_history_mode = False
            self._buffer_before_history = ""

    def _longest_common_prefix(self, strings):
        """Compute longest common prefix for a list of strings."""
        if not strings:
            return ""
        s1 = min(strings)
        s2 = max(strings)
        for i, c in enumerate(s1):
            if c != s2[i]:
                return s1[:i]
        return s1

    def _try_autocomplete(self) -> bool:
        """
        Try to auto-complete the current text based on supported commands.

        :return: True if something was completed, False otherwise
        """
        if not self._commands:
            return False

        text = self.text()
        # only complete first token; do not attempt when there is a space
        if not text or " " in text.strip():
            return False

        prefix = text.strip()
        # case-insensitive matching, but insert canonical command form
        matches = [cmd for cmd in self._commands if cmd.lower().startswith(prefix.lower())]

        if len(matches) == 1:
            self.setText(matches[0])
            self.end(False)
            return True

        if len(matches) > 1:
            # complete to the longest common prefix if it extends current prefix
            lcp = self._longest_common_prefix(matches)
            if len(lcp) > len(prefix):
                self.setText(lcp)
                self.end(False)
                return True

        return False

    def keyPressEvent(self, event):
        """
        Key press event

        :param event: key event
        """
        key = event.key()

        # Enter/Return sends the command
        if key in (QtCore.Qt.Key_Return, QtCore.Qt.Key_Enter):
            self.window.core.debug.console.on_send()
            self.setFocus()
            event.accept()
            return

        # Up/Down for history navigation
        if key == QtCore.Qt.Key_Up:
            self._history_prev()
            event.accept()
            return

        if key == QtCore.Qt.Key_Down:
            self._history_next()
            event.accept()
            return

        # TAB or SHIFT+TAB for command auto-completion
        if key in (QtCore.Qt.Key_Tab, QtCore.Qt.Key_Backtab):
            completed = self._try_autocomplete()
            if not completed:
                QApplication.beep()
            event.accept()  # prevent focus traversal
            return

        # Any other key exits history navigation mode
        if self._in_history_mode:
            self._in_history_mode = False
            self._history_index = len(self._history)
            self._buffer_before_history = ""

        super(ConsoleInput, self).keyPressEvent(event)

    def focusNextPrevChild(self, next):
        """
        Disable focus traversal with TAB for this widget.
        This prevents Qt from moving focus to other widgets when TAB/SHIFT+TAB is pressed.
        """
        return False


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.21 01:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt, QEvent, QTimer, QObject
from PySide6.QtWidgets import QScrollArea, QSplitter, QSplitterHandle, QPushButton, \
    QCheckBox, QHBoxLayout, QAbstractItemView, QLabel, QVBoxLayout

from pygpt_net.ui.widget.dialog.debug import DebugDialog
from pygpt_net.ui.widget.lists.debug import DebugList
from pygpt_net.ui.widget.textarea.editor import CodeEditor


class SmoothSplitterHandle(QSplitterHandle):
    """
    Splitter handle that notifies the parent splitter about drag begin/end.
    This allows the splitter to temporarily freeze heavy widgets during drag.
    """
    def __init__(self, orientation, parent):
        super().__init__(orientation, parent)
        self._drag_active = False

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            self._drag_active = True
            sp = self.splitter()
            if hasattr(sp, "_on_drag_begin"):
                sp._on_drag_begin()
        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        super().mouseReleaseEvent(event)
        if self._drag_active:
            sp = self.splitter()
            if hasattr(sp, "_on_drag_end"):
                sp._on_drag_end()
            self._drag_active = False


class SmoothSplitter(QSplitter):
    """
    Splitter optimized for heavy child widgets:
    - Uses non-opaque resize to avoid continuous relayout during drag.
    - Freezes updates on registered heavy widgets while dragging.
    """
    def __init__(self, orientation, parent=None):
        super().__init__(orientation, parent)
        self._heavy_widgets = []
        self.setOpaqueResize(False)            # rubber-band; resize applied on release
        self.setChildrenCollapsible(False)     # keep both panes visible

    def createHandle(self):
        return SmoothSplitterHandle(self.orientation(), self)

    def add_heavy_widget(self, w):
        if w and w not in self._heavy_widgets:
            self._heavy_widgets.append(w)

    def _on_drag_begin(self):
        for w in self._heavy_widgets:
            try:
                w.setUpdatesEnabled(False)
            except Exception:
                pass

    def _on_drag_end(self):
        for w in self._heavy_widgets:
            try:
                w.setUpdatesEnabled(True)
                w.update()
            except Exception:
                pass


class ResizeFreezeFilter(QObject):
    """
    Event filter that throttles expensive repaints during parent widget resize.
    Freezes updates for registered heavy widgets while the resize is in progress,
    and restores them shortly after the last resize event.
    """
    def __init__(self, heavy_widgets, parent=None, defer_ms=120):
        super().__init__(parent)
        self._heavy_widgets = [w for w in heavy_widgets if w is not None]
        self._defer_ms = defer_ms
        self._timer = QTimer(self)
        self._timer.setSingleShot(True)
        self._timer.timeout.connect(self._unfreeze)
        self._frozen = False
        self._saved_view_update_modes = {}

    def eventFilter(self, obj, event):
        if event.type() == QEvent.Resize:
            self._freeze()
            self._timer.start(self._defer_ms)
        return False

    def _freeze(self):
        if self._frozen:
            return
        self._frozen = True
        for w in self._heavy_widgets:
            try:
                if isinstance(w, QAbstractItemView):
                    self._saved_view_update_modes[w] = w.viewportUpdateMode()
                    w.setViewportUpdateMode(QAbstractItemView.NoViewportUpdate)
            except Exception:
                pass
            try:
                w.setUpdatesEnabled(False)
            except Exception:
                pass

    def _unfreeze(self):
        if not self._frozen:
            return
        self._frozen = False
        for w in self._heavy_widgets:
            try:
                w.setUpdatesEnabled(True)
            except Exception:
                pass
            try:
                if isinstance(w, QAbstractItemView) and w in self._saved_view_update_modes:
                    w.setViewportUpdateMode(self._saved_view_update_modes.pop(w))
            except Exception:
                pass
            try:
                w.update()
            except Exception:
                pass


class Debug:
    def __init__(self, window=None):
        """
        Debug setup

        :param window: Window instance
        """
        self.window = window

    def setup(self, id: str):
        """
        Setup debug dialog

        :param id: debug id
        """
        self.window.ui.debug[id] = DebugList(self.window)

        scroll = QScrollArea()
        scroll.setWidget(self.window.ui.debug[id])
        scroll.setWidgetResizable(True)

        # data viewer
        viewer = CodeEditor(self.window)
        viewer.setReadOnly(True)
        self.window.ui.debug[id].viewer = viewer

        # optimized splitter for heavy widgets
        splitter = SmoothSplitter(Qt.Horizontal)
        splitter.addWidget(scroll)
        splitter.addWidget(viewer)
        splitter.setStretchFactor(0, 3)
        splitter.setStretchFactor(1, 1)

        # register heavy widgets to freeze during drag
        splitter.add_heavy_widget(self.window.ui.debug[id])
        splitter.add_heavy_widget(self.window.ui.debug[id].viewport())
        splitter.add_heavy_widget(viewer)

        realtime_btn = QCheckBox("Realtime")
        realtime_btn.setChecked(False)
        realtime_btn.toggled.connect(
            lambda checked: self.window.controller.dialogs.debug.set_realtime(id, checked)
        )

        refresh_btn = QPushButton("Refresh")
        refresh_btn.clicked.connect(
            lambda: self.window.controller.dialogs.debug.refresh(id)
        )

        last_update_label = QLabel("Last update: N/A")
        self.window.ui.debug[id].last_update_label = last_update_label

        opts_layout = QHBoxLayout()
        opts_layout.addWidget(realtime_btn)
        opts_layout.addStretch(1)
        opts_layout.addWidget(refresh_btn)
        opts_layout.addStretch(1)
        opts_layout.addWidget(last_update_label)

        layout = QVBoxLayout()
        layout.addWidget(splitter, 1)
        layout.addLayout(opts_layout, 0)
        layout.setContentsMargins(5, 5, 5, 5)

        self.window.ui.dialog['debug.' + id] = DebugDialog(self.window, id)
        self.window.ui.dialog['debug.' + id].setLayout(layout)
        self.window.ui.dialog['debug.' + id].setWindowTitle("Debug" + ": " + id)

        # throttle heavy repaints during dialog resize
        heavy_widgets = [
            self.window.ui.debug[id],
            getattr(self.window.ui.debug[id], "viewport", lambda: None)(),
            viewer,
            scroll,
            getattr(scroll, "viewport", lambda: None)(),
        ]
        resize_filter = ResizeFreezeFilter(heavy_widgets, parent=self.window.ui.dialog['debug.' + id], defer_ms=120)
        self.window.ui.dialog['debug.' + id].installEventFilter(resize_filter)
        splitter.installEventFilter(resize_filter)
        # keep a strong reference to avoid garbage collection
        self.window.ui.dialog['debug.' + id]._resize_freeze_filter = resize_filter


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.21 01:00:00                  #
# ================================================== #

import time
from typing import Any

from PySide6.QtCore import Qt
from PySide6.QtGui import QStandardItemModel

from pygpt_net.core.debug.agent import AgentDebug
from pygpt_net.core.debug.agent_builder import AgentBuilderDebug
from pygpt_net.core.debug.assistants import AssistantsDebug
from pygpt_net.core.debug.attachments import AttachmentsDebug
from pygpt_net.core.debug.config import ConfigDebug
from pygpt_net.core.debug.context import ContextDebug
from pygpt_net.core.debug.db import DatabaseDebug
from pygpt_net.core.debug.events import EventsDebug
from pygpt_net.core.debug.indexes import IndexesDebug
from pygpt_net.core.debug.kernel import KernelDebug
from pygpt_net.core.debug.models import ModelsDebug
from pygpt_net.core.debug.plugins import PluginsDebug
from pygpt_net.core.debug.presets import PresetsDebug
from pygpt_net.core.debug.tabs import TabsDebug
from pygpt_net.core.debug.ui import UIDebug


class Debug:

    DBG_KEY, DBG_VALUE = range(2)

    def __init__(self, window=None):
        """
        Debugger handler

        :param window: Window instance
        """
        self.window = window

        # setup workers
        self.workers = {
            'agent': AgentDebug(self.window),
            'agent_builder': AgentBuilderDebug(self.window),
            'assistants': AssistantsDebug(self.window),
            'attachments': AttachmentsDebug(self.window),
            'config': ConfigDebug(self.window),
            'context': ContextDebug(self.window),
            'db': DatabaseDebug(self.window),
            'events': EventsDebug(self.window),
            'indexes': IndexesDebug(self.window),
            'kernel': KernelDebug(self.window),
            'models': ModelsDebug(self.window),
            'plugins': PluginsDebug(self.window),
            'presets': PresetsDebug(self.window),
            'tabs': TabsDebug(self.window),
            'ui': UIDebug(self.window)
        }

        self.is_realtime = False  # real-time data update

        # prepare debug ids
        self.ids = self.workers.keys()
        self.models = {}
        self.initialized = {}
        self.active = {}
        self.idx = {}
        self.counters = {}

        # prepare debug workers data
        for id in self.ids:
            self.models[id] = self.create_model(self.window)
            self.initialized[id] = False
            self.active[id] = False
            self.idx[id] = 0

    def refresh(self, id: str):
        """
        Refresh debug data

        :param id: debug id
        """
        if id in self.initialized:
            self.initialized[id] = False

    def begin(self, id: str):
        """
        Begin debug data

        :param id: debug id
        """
        if self.initialized[id] and not self.is_realtime:
            if self.is_active(id):
                return  # skip if already initialized and realtime disabled

        upd_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        self.window.ui.debug[id].last_update_label.setText(
            f"Last update: {upd_time}"
        )

        model = self.models.get(id, None)
        dialog = self.window.ui.debug[id]
        dialog.setModel(model)
        model.dataChanged.connect(dialog.on_data_begin)

        dialog.setUpdatesEnabled(False)
        if id not in self.counters or self.counters[id] != model.rowCount():
            model.removeRows(0, model.rowCount())
            self.initialized[id] = False
        self.idx[id] = 0

    def set_realtime(self, id: str, enabled: bool):
        """
        Set realtime mode

        :param id: debug id
        :param enabled: enable realtime
        """
        self.is_realtime = enabled

    def end(self, id: str):
        """
        End debug data

        :param id: debug id
        """
        dialog = self.window.ui.debug[id]
        self.counters[id] = self.idx[id]
        self.initialized[id] = True
        self.window.ui.debug[id].on_data_end()
        dialog.setUpdatesEnabled(True)

    def add(self, id: str, k: str, v: Any):
        """
        Append debug entry

        :param id: debug id
        :param k: key
        :param v: value
        """
        model = self.models.get(id, None)
        if self.initialized[id] is False:
            idx = model.rowCount()
            model.insertRow(idx)
            model.setData(model.index(idx, self.DBG_KEY), k)
            model.setData(model.index(idx, self.DBG_VALUE), v)
            self.idx[id] += 1
        else:
            for idx in range(0, model.rowCount()):
                self.idx[id] += 1
                if model.index(idx, self.DBG_KEY).data() == k:
                    if model.index(idx, self.DBG_VALUE).data() == v:
                        continue  # no changes, skip update
                    model.setData(model.index(idx, self.DBG_VALUE), v)
                    return

    def get_ids(self) -> list:
        """
        Get debug workers ids

        :return: list of ids
        """
        return list(self.ids)

    def get_workers(self) -> dict:
        """
        Get debug workers

        :return: dict of workers
        """
        return self.workers

    def update_worker(self, id: str):
        """
        Update debug worker

        :param id: debug id
        """
        if id in self.workers:
            self.workers[id].update()

    def is_active(self, id: str) -> bool:
        """
        Check if debug window is active

        :param id: debug id
        :return: True if active
        """
        if id not in self.active:
            return False
        return self.active[id]

    def show(self, id: str):
        """
        Activate debug window

        :param id: debug id
        """
        if id not in self.active:
            return
        self.active[id] = True
        self.window.ui.dialogs.open(f"debug.{id}", width=800, height=600)

    def hide(self, id: str):
        """
        Deactivate debug window

        :param id: debug id
        """
        if id not in self.active:
            return
        self.active[id] = False
        self.window.ui.dialogs.close(f"debug.{id}")

    def create_model(self, parent) -> QStandardItemModel:
        """
        Create list model

        :param parent: parent widget
        :return: model instance
        """
        model = QStandardItemModel(0, 2, parent)
        model.setHeaderData(self.DBG_KEY, Qt.Horizontal, "Key")
        model.setHeaderData(self.DBG_VALUE, Qt.Horizontal, "Value")
        return model
